(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/installer.js                                                 //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Fs = Npm.require('fs');                                            // 17
var Path = Npm.require('path');                                        // 18
var Crypto = Npm.require('crypto');                                    // 19
var ChildProcess = Npm.require('child_process');                       // 20
var Http = Npm.require('http');                                        // 21
var Https = Npm.require('https');                                      // 22
var Url = Npm.require('url');                                          // 23
var Promise = Npm.require('es6-promise').Promise;                      // 24
var Capnp = Npm.require('capnp');                                      // 25
                                                                       //
var Manifest = Capnp.importSystem('sandstorm/package.capnp').Manifest;
                                                                       //
var installers = undefined; // set to {} on main replica               // 29
// To protect against race conditions, we require that each row in the Packages
// collection have at most one writer at a time, as tracked by this `installers`
// map. Each key is a package ID and each value is either an AppInstaller object
// or the string 'uninstalling', indicating that some fiber is working on
// uninstalling the package.                                           //
                                                                       //
var verifyIsMainReplica = function () {                                // 36
  if (Meteor.settings.replicaNumber) {                                 // 37
    throw new Error('This can only be called on the main front-end replica.');
  }                                                                    //
};                                                                     //
                                                                       //
Meteor.methods({                                                       // 42
  deleteUnusedPackages: function (appId) {                             // 43
    check(appId, String);                                              // 44
    Packages.find({ appId: appId }).forEach(function (pkg) {           // 45
      deletePackage(pkg._id);                                          // 45
    });                                                                //
  }                                                                    //
});                                                                    //
                                                                       //
deletePackage = function (packageId) {                                 // 49
  // Mark package for possible deletion;                               //
  Packages.update({ _id: packageId, status: 'ready' }, { $set: { shouldCleanup: true } });
};                                                                     //
                                                                       //
var deletePackageInternal = function (pkg) {                           // 54
  verifyIsMainReplica();                                               // 55
                                                                       //
  var packageId = pkg._id;                                             // 57
                                                                       //
  if (packageId in installers) {                                       // 59
    return;                                                            // 60
  }                                                                    //
                                                                       //
  installers[packageId] = 'uninstalling';                              // 63
                                                                       //
  try {                                                                // 65
    var action = UserActions.findOne({ packageId: packageId });        // 66
    var grain = Grains.findOne({ packageId: packageId });              // 67
    var notificationQuery = {};                                        // 68
    notificationQuery['appUpdates.' + pkg.appId] = { $exists: true };  // 69
    if (!grain && !action && !(pkg.isAutoUpdated && Notifications.findOne(notificationQuery))) {
      Packages.update({ _id: packageId }, { $set: { status: 'delete' }, $unset: { shouldCleanup: '' } });
      waitPromise(globalBackend.cap().deletePackage(packageId));       // 72
      Packages.remove(packageId);                                      // 73
                                                                       //
      // Clean up assets (icon, etc).                                  //
      getAllManifestAssets(pkg.manifest).forEach(function (assetId) {  // 76
        globalDb.unrefStaticAsset(assetId);                            // 77
      });                                                              //
    } else {                                                           //
      Packages.update({ _id: packageId }, { $unset: { shouldCleanup: '' } });
    }                                                                  //
                                                                       //
    delete installers[packageId];                                      // 83
  } catch (error) {                                                    //
    delete installers[packageId];                                      // 85
    throw error;                                                       // 86
  }                                                                    //
};                                                                     //
                                                                       //
var startInstallInternal = function (pkg) {                            // 90
  verifyIsMainReplica();                                               // 91
                                                                       //
  if (pkg._id in installers) {                                         // 93
    return;                                                            // 94
  }                                                                    //
                                                                       //
  var installer = new AppInstaller(pkg._id, pkg.url, pkg.appId, pkg.isAutoUpdated);
  installers[pkg._id] = installer;                                     // 98
  installer.start();                                                   // 99
};                                                                     //
                                                                       //
cancelDownload = function (packageId) {                                // 102
  Packages.remove({ _id: packageId, status: 'download' });             // 103
};                                                                     //
                                                                       //
var cancelDownloadInternal = function (pkg) {                          // 106
  verifyIsMainReplica();                                               // 107
                                                                       //
  var installer = installers[pkg._id];                                 // 109
                                                                       //
  // Don't do anything unless a download is in progress.               //
  if (installer && installer.downloadRequest) {                        // 112
    // OK, effect cancellation by faking an error.                     //
    installer.wrapCallback(function () {                               // 114
      throw new Error('Canceled');                                     // 115
    })();                                                              //
  }                                                                    //
};                                                                     //
                                                                       //
if (!Meteor.settings.replicaNumber) {                                  // 120
  installers = {};                                                     // 121
                                                                       //
  Meteor.startup(function () {                                         // 123
    // Restart any deletions that were killed while in-progress.       //
    Packages.find({ status: 'delete' }).forEach(deletePackageInternal);
                                                                       //
    // Watch for new installation requests and fulfill them.           //
    Packages.find({ status: { $in: ['download', 'verify', 'unpack', 'analyze'] } }).observe({
      added: startInstallInternal,                                     // 129
      removed: cancelDownloadInternal                                  // 130
    });                                                                //
                                                                       //
    // Watch for new cleanup requests and fulfill them.                //
    Packages.find({ status: 'ready', shouldCleanup: true }).observe({  // 134
      added: deletePackageInternal                                     // 135
    });                                                                //
  });                                                                  //
}                                                                      //
                                                                       //
doClientUpload = function (stream) {                                   // 140
  return new Promise(function (resolve, reject) {                      // 141
    var id = Random.id();                                              // 142
                                                                       //
    var backendStream = globalBackend.cap().installPackage().stream;   // 144
    var hasher = Crypto.createHash('sha256');                          // 145
                                                                       //
    stream.on('data', function (chunk) {                               // 147
      try {                                                            // 148
        hasher.update(chunk);                                          // 149
        backendStream.write(chunk);                                    // 150
      } catch (err) {                                                  //
        reject(err);                                                   // 152
      }                                                                //
    });                                                                //
                                                                       //
    stream.on('end', function () {                                     // 156
      try {                                                            // 157
        (function () {                                                 //
          backendStream.done();                                        // 158
          var packageId = hasher.digest('hex').slice(0, 32);           // 159
          resolve(backendStream.saveAs(packageId).then(function () {   // 160
            return packageId;                                          // 161
          }));                                                         //
          backendStream.close();                                       // 163
        })();                                                          //
      } catch (err) {                                                  //
        reject(err);                                                   // 165
      }                                                                //
    });                                                                //
                                                                       //
    stream.on('error', function (err) {                                // 169
      // TODO(soon):  This event does't seem to fire if the user leaves the page mid-upload.
      try {                                                            // 171
        backendStream.close();                                         // 172
        reject(err);                                                   // 173
      } catch (err2) {                                                 //
        reject(err2);                                                  // 175
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
AppInstaller = (function () {                                          // 181
  function AppInstaller(packageId, url, appId, isAutoUpdated) {        // 182
    babelHelpers.classCallCheck(this, AppInstaller);                   //
                                                                       //
    verifyIsMainReplica();                                             // 183
                                                                       //
    this.packageId = packageId;                                        // 185
    this.url = url;                                                    // 186
    this.failed = false;                                               // 187
    this.appId = appId;                                                // 188
    this.isAutoUpdated = isAutoUpdated;                                // 189
                                                                       //
    // Serializes database writes.                                     //
    this.writeChain = Promise.resolve();                               // 192
  }                                                                    //
                                                                       //
  AppInstaller.prototype.updateProgress = (function () {               // 181
    function updateProgress(status, progress, error, manifest) {       // 195
      // TODO(security):  On error, we should actually delete the package from the database and only
      //   display the error to whomever was watching at the time.  Otherwise it's easy to confuse
      //   people by 'pre-failing' packages.  (Actually, perhaps if a user tries to download an
      //   already-downloading package but specifies a different URL, we really should initiate an
      //   entirely separate download...  but cancel it if the first download succeeds.)
                                                                       //
      this.status = status;                                            // 202
      this.progress = progress || -1;                                  // 203
      this.error = error;                                              // 204
      this.manifest = manifest || null;                                // 205
                                                                       //
      var _this = this;                                                // 207
                                                                       //
      // The callback passed to inMeteor() runs in a new fiber. We need to make sure database writes
      // occur in exactly the order in which we generate them, so we use a promise chain to serialize
      // them.                                                         //
      this.writeChain = this.writeChain.then(function () {             // 212
        return inMeteor(function () {                                  // 213
          if (manifest) extractManifestAssets(manifest);               // 214
                                                                       //
          Packages.update(_this.packageId, {                           // 216
            $set: {                                                    // 217
              status: _this.status,                                    // 218
              progress: _this.progress,                                // 219
              error: _this.error ? _this.error.message : null,         // 220
              manifest: _this.manifest,                                // 221
              appId: _this.appId,                                      // 222
              authorPgpKeyFingerprint: _this.authorPgpKeyFingerprint   // 223
            }                                                          //
          });                                                          //
                                                                       //
          if (_this.authorPgpKeyFingerprint) {                         // 227
            globalDb.updateKeybaseProfileAsync(_this.authorPgpKeyFingerprint);
          }                                                            //
        })['catch'](function (err) {                                   //
          console.error(err.stack);                                    // 231
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return updateProgress;                                             //
  })();                                                                //
                                                                       //
  AppInstaller.prototype.wrapCallback = (function () {                 // 181
    function wrapCallback(method) {                                    // 236
      // Note that the function below must not be an arrow function, since arrow functions do not have
      // access to the context's arguments array.                      //
      var _this = this;                                                // 239
      return function () {                                             // 240
        if (_this.failed) return;                                      // 241
        try {                                                          // 242
          return method.apply(_this, _.toArray(arguments));            // 243
        } catch (err) {                                                //
          _this.failed = true;                                         // 245
          _this.cleanup();                                             // 246
          _this.updateProgress('failed', 0, err);                      // 247
          _this.writeChain = _this.writeChain.then(function () {       // 248
            delete installers[_this.packageId];                        // 249
          });                                                          //
          console.error('Failed to install app:', err.stack);          // 251
        }                                                              //
      };                                                               //
    }                                                                  //
                                                                       //
    return wrapCallback;                                               //
  })();                                                                //
                                                                       //
  AppInstaller.prototype.cleanup = (function () {                      // 181
    function cleanup() {                                               // 256
      if (this.uploadStream) {                                         // 257
        try {                                                          // 258
          this.uploadStream.close();                                   // 259
        } catch (err) {}                                               //
                                                                       //
        delete this.uploadStream;                                      // 262
      }                                                                //
                                                                       //
      if (this.downloadRequest) {                                      // 265
        try {                                                          // 266
          this.downloadRequest.abort();                                // 267
        } catch (err) {}                                               //
                                                                       //
        delete this.downloadRequest;                                   // 270
      }                                                                //
    }                                                                  //
                                                                       //
    return cleanup;                                                    //
  })();                                                                //
                                                                       //
  AppInstaller.prototype.start = (function () {                        // 181
    function start() {                                                 // 274
      var _this2 = this;                                               //
                                                                       //
      return this.wrapCallback(function () {                           // 275
        _this2.cleanup();                                              // 276
                                                                       //
        globalBackend.cap().tryGetPackage(_this2.packageId).then(_this2.wrapCallback(function (info) {
          if (info.appId) {                                            // 279
            _this2.appId = info.appId;                                 // 280
            _this2.authorPgpKeyFingerprint = info.authorPgpKeyFingerprint;
            _this2.done(info.manifest);                                // 282
          } else {                                                     //
            _this2.doDownload();                                       // 284
          }                                                            //
        }), _this2.wrapCallback(function (err) {                       //
          throw err;                                                   // 287
        }));                                                           //
      })();                                                            //
    }                                                                  //
                                                                       //
    return start;                                                      //
  })();                                                                //
                                                                       //
  AppInstaller.prototype.doDownload = (function () {                   // 181
    function doDownload() {                                            // 292
      if (!this.url) {                                                 // 293
        throw new Error('Unknown package ID, and no URL was provided.');
      }                                                                //
                                                                       //
      console.log('Downloading app:', this.url);                       // 297
      this.updateProgress('download');                                 // 298
                                                                       //
      this.uploadStream = globalBackend.cap().installPackage().stream;
      return this.doDownloadTo(this.uploadStream);                     // 301
    }                                                                  //
                                                                       //
    return doDownload;                                                 //
  })();                                                                //
                                                                       //
  AppInstaller.prototype.doDownloadTo = (function () {                 // 181
    function doDownloadTo(out) {                                       // 304
      var _this3 = this;                                               //
                                                                       //
      var url = Url.parse(this.url);                                   // 305
      var options = {                                                  // 306
        hostname: url.hostname,                                        // 307
        port: url.port,                                                // 308
        path: url.path                                                 // 309
      };                                                               //
                                                                       //
      var protocol = undefined;                                        // 312
      if (url.protocol === 'http:' || url.protocol === 'https:') {     // 313
        protocol = Request.defaults({                                  // 314
          maxRedirects: 20,                                            // 315
          // Since we will verify the download against a hash anyway, we don't need to verify the
          // server's certificate. In fact, the only reason we support HTTPS at all here is because
          // some servers refuse to serve over HTTP (which is, in general, a good thing). Skipping the
          // certificate check here is helpful in that it means we don't have to worry about having a
          // reasonable list of trusted CAs available to Sandstorm.    //
          strictSSL: false                                             // 321
        });                                                            //
      } else {                                                         //
        throw new Error('Protocol not supported: ' + url.protocol);    // 324
      }                                                                //
                                                                       //
      // TODO(security):  It could arguably be a security problem that it's possible to probe the
      //   server's local network (behind any firewalls) by presenting URLs here.
      var bytesExpected = undefined;                                   // 329
      var bytesReceived = 0;                                           // 330
      var hasher = Crypto.createHash('sha256');                        // 331
      var done = false;                                                // 332
      var updateDownloadProgress = _.throttle(this.wrapCallback(function () {
        if (!done) {                                                   // 334
          if (bytesExpected) {                                         // 335
            _this3.updateProgress('download', bytesReceived / bytesExpected);
          } else {                                                     //
            _this3.updateProgress('download', bytesReceived);          // 338
          }                                                            //
        }                                                              //
      }), 500);                                                        //
                                                                       //
      var request = protocol.get(this.url);                            // 343
                                                                       //
      request.on('response', this.wrapCallback(function (response) {   // 345
        if ('content-length' in response.headers) {                    // 346
          bytesExpected = parseInt(response.headers['content-length']);
        }                                                              //
      }));                                                             //
                                                                       //
      request.on('data', this.wrapCallback(function (chunk) {          // 351
        hasher.update(chunk);                                          // 352
        out.write(chunk);                                              // 353
        bytesReceived += chunk.length;                                 // 354
        updateDownloadProgress();                                      // 355
      }));                                                             //
                                                                       //
      request.on('end', this.wrapCallback(function () {                // 358
        out.done();                                                    // 359
                                                                       //
        if (hasher.digest('hex').slice(0, 32) !== _this3.packageId) {  // 361
          throw new Error('Package hash did not match.');              // 362
        }                                                              //
                                                                       //
        done = true;                                                   // 365
        delete _this3.downloadRequest;                                 // 366
                                                                       //
        _this3.updateProgress('unpack');                               // 368
        out.saveAs(_this3.packageId).then(_this3.wrapCallback(function (info) {
          _this3.appId = info.appId;                                   // 370
          _this3.authorPgpKeyFingerprint = info.authorPgpKeyFingerprint;
          _this3.done(info.manifest);                                  // 372
        }), _this3.wrapCallback(function (err) {                       //
          throw err;                                                   // 374
        }));                                                           //
      }));                                                             //
                                                                       //
      request.on('error', this.wrapCallback(function (err) {           // 378
        throw err;                                                     // 378
      }));                                                             //
                                                                       //
      this.downloadRequest = request;                                  // 380
    }                                                                  //
                                                                       //
    return doDownloadTo;                                               //
  })();                                                                //
                                                                       //
  AppInstaller.prototype.done = (function () {                         // 181
    function done(manifest) {                                          // 383
      console.log('App ready:', this.packageId);                       // 384
      this.updateProgress('ready', 1, undefined, manifest);            // 385
      var _this = this;                                                // 386
      _this.writeChain = _this.writeChain.then(function () {           // 387
        return inMeteor(function () {                                  // 388
          if (_this.isAutoUpdated) {                                   // 389
            globalDb.sendAppUpdateNotifications(_this.appId, _this.packageId, manifest.appTitle && manifest.appTitle.defaultText, manifest.appVersion, manifest.appMarketingVersion && manifest.appMarketingVersion.defaultText);
          }                                                            //
        });                                                            //
      }).then(function () {                                            //
        delete installers[_this.packageId];                            // 396
      });                                                              //
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  return AppInstaller;                                                 //
})();                                                                  //
                                                                       //
extractManifestAssets = function (manifest) {                          // 401
  var metadata = manifest.metadata;                                    // 402
  if (!metadata) return;                                               // 403
                                                                       //
  var icons = metadata.icons;                                          // 405
  if (icons) {                                                         // 406
    var handleIcon = function (icon) {                                 // 407
      if (icon.svg) {                                                  // 408
        icon.assetId = globalDb.addStaticAsset({ mimeType: 'image/svg+xml' }, icon.svg);
        icon.format = 'svg';                                           // 410
        delete icon.svg;                                               // 411
        return true;                                                   // 412
      } else if (icon.png) {                                           //
        // Use the 1x version for 'normal' DPI, unless 1x isn't provided, in which case use 2x.
        var normalDpi = icon.png.dpi1x || icon.png.dpi2x;              // 415
        if (!normalDpi) return false;                                  // 416
        icon.format = 'png';                                           // 417
        icon.assetId = globalDb.addStaticAsset({ mimeType: 'image/png' }, normalDpi);
                                                                       //
        if (icon.png.dpi1x && icon.png.dpi2x) {                        // 420
          // Icon specifies both resolutions, so also record a 2x DPI option.
          icon.assetId2xDpi = globalDb.addStaticAsset({ mimeType: 'image/png' }, icon.png.dpi2x);
        }                                                              //
                                                                       //
        delete icon.png;                                               // 425
        return true;                                                   // 426
      } else {                                                         //
        // Unknown icon. Filter it.                                    //
        return false;                                                  // 429
      }                                                                //
    };                                                                 //
                                                                       //
    if (icons.appGrid && !handleIcon(icons.appGrid)) delete icons.appGrid;
    if (icons.grain && !handleIcon(icons.grain)) delete icons.grain;   // 434
                                                                       //
    // We don't need the market icons.                                 //
    if (icons.market) delete icons.market;                             // 437
    if (icons.marketBig) delete icons.marketBig;                       // 438
  }                                                                    //
                                                                       //
  var handleLocalizedText = function (text) {                          // 441
    if (text.defaultText) {                                            // 442
      text.defaultTextAssetId = globalDb.addStaticAsset({ mimeType: 'text/plain' }, text.defaultText);
      delete text.defaultText;                                         // 444
    }                                                                  //
                                                                       //
    if (text.localizations) {                                          // 447
      text.localizations.forEach(function (localization) {             // 448
        if (localization.text) {                                       // 449
          localization.assetId = globalDb.addStaticAsset({ mimeType: 'text/plain' }, localization.text);
          delete localization.text;                                    // 452
        }                                                              //
      });                                                              //
    }                                                                  //
  };                                                                   //
                                                                       //
  var license = metadata.license;                                      // 458
  if (license) {                                                       // 459
    if (license.proprietary) license.proprietary = handleLocalizedText(license.proprietary);
    if (license.publicDomain) license.publicDomain = handleLocalizedText(license.proprietary);
    if (license.notices) license.notices = handleLocalizedText(license.notices);
  }                                                                    //
                                                                       //
  var author = metadata.author;                                        // 465
  if (author) {                                                        // 466
    // We remove the PGP signature since it was already verified down to a key ID in the back-end.
    if (author.pgpSignature) delete author.pgpSignature;               // 468
  }                                                                    //
                                                                       //
  // Don't need the keyring either.                                    //
  if (metadata.pgpKeyring) delete metadata.pgpKeyring;                 // 472
                                                                       //
  // Perhaps used by the 'about' page?                                 //
  if (metadata.description) metadata.description = handleLocalizedText(metadata.description);
                                                                       //
  // Screenshots are for app marketing; we don't use them post-install.
  if (metadata.screenshots) delete metadata.screenshots;               // 478
                                                                       //
  // We might allow the user to view the changelog.                    //
  if (metadata.changeLog) metadata.changeLog = handleLocalizedText(metadata.changeLog);
};                                                                     //
                                                                       //
getAllManifestAssets = function (manifest) {                           // 484
  // Returns a list of all asset IDs in the given manifest.            //
                                                                       //
  var metadata = manifest.metadata;                                    // 487
  if (!metadata) return [];                                            // 488
                                                                       //
  var result = [];                                                     // 490
                                                                       //
  var icons = metadata.icons;                                          // 492
  if (icons) {                                                         // 493
    var handleIcon = function (icon) {                                 // 494
      if (icon.assetId) {                                              // 495
        result.push(icon.assetId);                                     // 496
      }                                                                //
    };                                                                 //
                                                                       //
    if (icons.appGrid) handleIcon(icons.appGrid);                      // 500
    if (icons.grain) handleIcon(icons.grain);                          // 501
  }                                                                    //
                                                                       //
  var handleLocalizedText = function (text) {                          // 504
    if (text.defaultTextAssetId) {                                     // 505
      result.push(defaultTextAssetId);                                 // 506
    }                                                                  //
                                                                       //
    if (text.localizations) {                                          // 509
      text.localizations.forEach(function (localization) {             // 510
        if (localization.assetId) {                                    // 511
          result.push(localization.assetId);                           // 512
        }                                                              //
      });                                                              //
    }                                                                  //
  };                                                                   //
                                                                       //
  var license = metadata.license;                                      // 518
  if (license) {                                                       // 519
    if (license.proprietary) handleLocalizedText(license.proprietary);
    if (license.publicDomain) handleLocalizedText(license.publicDomain);
    if (license.notices) handleLocalizedText(license.notices);         // 522
  }                                                                    //
                                                                       //
  if (metadata.description) handleLocalizedText(metadata.description);
  if (metadata.changeLog) handleLocalizedText(metadata.changeLog);     // 526
                                                                       //
  return result;                                                       // 528
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=installer.js.map
